'9/oct/2013 20:03
'----------------------------------------------------------------
'Peak Power Tracking Solar Charger for Maximite Computer
'Based on Arduino Peak Power Tracking Solar Charger by Tim Nolan
'(www.timnolan.com)   5/1/09
'----------------------------------------------------------------
'CONSTANTS:
SOL_AMPS_CHAN=1         'adc input for solar amps
SOL_V_CHAN=3            'adc input for solar volts
bat_v_CHAN=2            'adc input for battery volts
AVG_NUM=4       'number of adc readings to average. sampled every 5ms for
                '50Hz cancelling (best use even no. )
PWM_PIN=9               'the output pin for the pwm
PWM_ENABLE_PIN=11       'controls shutoff of the IR2184 MOSFET driver
PWM_MAX=99              'maximum pwm duty cyle %
PWM_MIN=40              'minimum pwm duty cyle %
PWM_START=62            'initial pwm duty cyle %
DELTA=0.25              'step size of the pwm value for the ppt algorithm
LOW_SOL_WATTS=2
MIN_SOL_WATTS=1
MIN_bat_v = 22.0
BAT_FLT_V = 28.2
HIGH_bat_v= 26.0
OFF_NUM=50            'number of iterations of off charger state
FREQ = 20100          'use 50(2n+1) ie xxx50 Hz when sampling at 10ms
                      'use 100(2n+1) ie xx100 Hz when sampling at 5ms

'**************************
'* START EXECUTION HERE:  *
'**************************
setup
'----------------------------------------------------------------
' Main loop.
'----------------------------------------------------------------
Do
  Pin(0)= y Mod 5       'This flashes the green led
  hour$=Mid$(Time$,1,2) 'find current hours
  If hour$<>old_hour$ Then  'if hours just changed then 
    'xxxxx this lot happens every hour  
    old_hour$=hour$
    If charger_state$ <> "off" Then   'dont write if near shutdown
      Pin(pwm_enable_pin)=0
      SaveBMP "b:\screens\scrn"+hour$+".bmp"
      Pause(2000)
      Cls
      plot_grid
      Pin(pwm_enable_pin)=1
    EndIf
    'xxxxx
  EndIf
  y=y+1
  print_data
  a$=Inkey$
  If a$="c" Then 'clear the screen
    Cls
    plot_grid
  ElseIf bat_v > 29.3 Then  'battery very high. Error?
    Pin(pwm_enable_pin)=0
    WatchDog off
    GoTo reboot
  ElseIf a$="d" Then  'display the stored screen grabs
    Pin(pwm_enable_pin)=0
    WatchDog off
    showscreens
  ElseIf a$<>"" Then  'any other key stops the program
    Pin(pwm_enable_pin)=0
    WatchDog off
    End
  EndIf
  If Timer < 1500 Then
    SetTick 0,0,2             'STOP run_charger For 1.5 sec
    Do While Timer < 1500
      Loop
    SetTick 50,run_charger,2
  EndIf
Loop

'------------------------------------------------------------------------------
'This routine reads and sums the analog input samples; solar volts, solar amps
'and battery volts. It is called every 10ms by timer interrupt. 4 samples over
'100ms are summed
'-------------------------------------------------------------------------------
Sub get_values
  z=z+1:If z=AVG_NUM Then z=0
  sv = Pin(3):sum_sv=sum_sv - sv(z) + sv : sv(z)=sv
  'read new value,subtract old value from sum,add new value to sum
  'then store new value
  bv = Pin(2):sum_bv=sum_bv - bv(z) + bv : bv(z)=bv
  sa = Pin(1):sum_sa=sum_sa - sa(z) + sa : sa(z)=sa
End Sub
'-----------------------------
Sub showscreens
Local k
k=0
Do While 0 <= k And k < 15
  LoadBMP "b:\screens\scrn"+Right$(Str$(k+100),2)+".bmp"
  Locate 0,0 : Print k : a$=Inkey$
  If a$ = "=" Or a$ = "+" Then  'use + key to advance
    k=k+1
  ElseIf a$ = "_" Or a$ = "-" Then  'use - key to reverse
    k=k-1
  EndIf
Loop
    Cls 'end if k<0 or k>15
    plot_grid
    WatchDog 2000 'reset if watchdog not reset for 2sec
End Sub

'---------------------------------------------------------------------------
' This routine is automatically called at powerup/reset
'---------------------------------------------------------------------------
Sub setup   ' run once, when the charger starts
' initialise global variables:
  Open "diagnose.txt" For random As #1
  Seek #1,1   ' no. of watchdog resets
  wdr = Val(Input$(5,#1))
  Seek #1,10  ' no. of overvoltage resets
  ovr = Val(Input$(5,#1)) 
  Close #1
  z=0:y=0:whrs2=0:whrs=0
  old_sol_watts = 0          'solar watts from previous time thru ppt routine
  old_hour$="00"
  charger_state$ = "bulk"        ' start with charger state as bulk
  Dim bv(AVG_NUM), sv(AVG_NUM), sa(AVG_NUM)
  Cls
  plot_grid
  SetPin PWM_ENABLE_PIN, 8     ' sets the digital pin as output
  SetPin SOL_AMPS_CHAN, 1      ' sets the ANALOG pin as INput
  SetPin SOL_V_CHAN, 1         ' sets the ANALOG pin as INput
  SetPin bat_v_CHAN, 1         ' sets the ANALOG pin as INput
  PWM FREQ,PWM_START           ' starting value for pwm
  pwm=PWM_START                '@@ ADDED 13/9/13
  SetTick 5,get_values,3  'sample a/d inputs every 5ms
    '10ms/n cancels 50hz interference
  SetTick 50,read_data,1
  Pause 500
  Pin(PWM_ENABLE_PIN)=1        ' turn on MOSFET driver chip
  SetTick 50,run_charger,2
End Sub

'----------------------------------------------------------------------
'This routine sets the pwm duty cycle.
'There is a special case for 100% duty cycle.
'Normally this would be have the top MOSFET on all the time but the
'MOSFET driver IR2184 uses a charge pump To generate the gate
'voltage so it has to keep running all the time. So for 100% duty cycle
'I set the pwm value To PWM_FULL so it is on 99%, almost full on but
'switches enough To keep the charge pump on IR2184 working.
'-----------------------------------------------------------------------
Sub set_pwm_duty
  If pwm > PWM_MAX Then
     pwm = PWM_MAX           'check limits of PWM duty cyle and set to PWM_MAX
  ElseIf pwm < PWM_MIN Then
     pwm = PWM_MIN           'if pwm is less than PWM_MIN then set to PWM_MIN
  EndIf
PWM FREQ,pwm
End Sub

'-----------------------------------------------------------------------
' This routine prints the data.
'-----------------------------------------------------------------------
Sub print_data
  Locate 0,60
  Print "ovr = ";ovr
  Print "wdr = ";wdr
  Print "time ";Time$
  Print " FREQ = ";FREQ;"  "
  if charger_state$ = "bulk" then
    Print " delta = ";delta
    else 
    print "                     "
  endif
  Print " charger = ";charger_state$;"  "
  Print
  Print "  SOLAR"
  Print "  -----"
  Print "       pwm = ";Format$(pwm,"%6.1f")
  Print "  Sol_Amps = ";Format$(sol_amps,"%6.2f")
  Print " Sol_Volts = ";Format$(SOL_V,"%6.2f")
  Print
  Print "  WATTS"
  Print "  -----"
  Print "  NOW      ";Format$(sol_watts,"%6.2f")
  Print "  MAXIMUM  ";Format$(max_watts,"%6.2f")
  Print "  Watt_hrs ";Format$(whrs+whrs2,"%6.3f")
  Print
  Print "  BAT_VOLTS"
  Print "  ---------"
  Print "    NOW     ";Format$(bat_v,"%6.2f")
  Print "    MAX     ";Format$(max_batv,"%6.2f");" ";mbv$;" "
  Print "    FLT_SET ";Format$(BAT_FLT_V,"%6.2f")
  Print
  Print "off_count= "; off_count ;" "
If sol_watts > max_watts Then max_watts = sol_watts
If bat_v > max_batv Then
  max_batv = bat_v
  mbv$=charger_state$
EndIf
Pixel(169+3*pwm,400-sol_watts/4)=1  'max val=4*432= 1728 watts
Pixel((y Mod 300) +170,400-bat_v*12.5)=1  '1000w=20v ie factor=50. 50/4=12.5
End Sub
'-----------------------------------------------
'this routine plots the grid and axes for the
'power vs pwm graph
'-----------------------------------------------
Sub plot_grid
Local p,w
For p=0 To 100 Step 10
For w=0 To 1600 Step 100
Pixel(169+3*p,400-w/4)=1  'max val=4*432= 1728 watts
If p=0 Then           'label vert axis
  Locate 135,395-w/4
  Print w
  EndIf
If w=0 Then
  Locate 155+3*p,405
  Print p
  EndIf
Next w
Next p
w=0
For p=10 To 50 Step 10
w=w+1
Locate 125,p : Print Mid$("POWER",w,1)
Next p
Locate 120,60 : Print "(W)"
Locate 190,390 : Print " PWM PERCENTAGE"
End Sub
'------------------------------------------------------------------------------
' This routine reads all the analog input values.
' It also calculates the input watts from solar amps * solar voltage
' actual volts=v.read * (r1+r2)/r2.
' for 24v batt we need max reading of 32v. Assuming 3.3v reference, factor
' needed = 9.697. I use 22k and 2k2 for factor of 11 giving full scale
' of 36.3v . We need max of 50v on solar side so factor needed=15.15
' I use 33k and 2k2 gives ratio of 16.00
'------------------------------------------------------------------------------
Sub read_data
  sol_amps =  sum_sa/AVG_NUM  / .001 / 60  '1mohm shunt, 60 amplification factor
    '3.3v /.001/60 = 55amp max..input
  SOL_V = sum_sv/AVG_NUM * 16 * 0.97
  bat_v = sum_bv/AVG_NUM * 11 * 0.982
  sol_watts = sol_amps * SOL_V
'use 2 variables for watthour sum to avoid floating point errors
'  whrs=whrs+.05*sol_watts/3600 'updated every 50ms. 3600=seconds per hour
  whrs2=whrs2+sol_watts/72000
  If whrs2 > 1 Then
    whrs=whrs+whrs2
    whrs2=0
  EndIf
End Sub

'------------------------------------------------------------------------------
' This routine is the charger state machine. It has four states on, off, bulk and float.
' It's called every 50ms by timer interrupt to see what state the charger should be in.
'
'On State - this is charger state for MIN_SOL_WATTS < solar watts < LOW_SOL_WATTS. This state is
'probably happening at dawn and dusk when the solar watts input is too low for the bulk charging
'state but not low enough to go into the off state. In this state we just set the pwm = 100% to
'get the most of low amount of power available.
'
'Bulk State - this is charger state for solar watts > MIN_SOL_WATTS. This is where we do the bulk of
'the battery charging and where we run the Peak Power Tracking alogorithm. In this state we try and
'run the maximum amount of power that the solar panels can generate into the battery.

'Float State - As the battery charges it's voltage rises. When it gets to the BAT_FLT_V we are
'done with the bulk battery charging and enter the battery float state. In this state we try to
'keep the battery voltage at BAT_FLT_V by adjusting the pwm value. If we get to pwm = 100%
'it means we can't keep the battery voltage at BAT_FLT_V which probably means the battery is
'being drawn down by some load so we need to back into the bulk charging mode.

'Off State - This is state that the charger enters when solar watts < MIN_SOL_WATTS. The charger
'goes into this state when it gets dark and there is no more power being generated by the solar
'panels. The MOSFETs are turned off in this state so that power from the battery doesn't leak back
'into the solar panel. When the charger off state is first entered all it does is decrement
'off_count for OFF_NUM times. This is done because if the battery is disconnected (or battery fuse
'is blown) it takes some time before the battery voltage changes enough so we can tell that
'the battery is no longer connected. This off_count gives some time for battery voltage to
'change so we can tell this.
'------------------------------------------------------------------------------
Sub run_charger
x=x+1   'record how many times the routine has run
WatchDog 2000
'----------
If charger_state$="on" Then
  If sol_watts < MIN_SOL_WATTS Then    'if watts input from the solar panel is less than
     charger_state$ = "off"            'the minimum solar watts then it is gettig dark so
     off_count = OFF_NUM               'go to the charger off state
     Pin(PWM_ENABLE_PIN)=0
  ElseIf bat_v > BAT_FLT_V Then        'else if the battery voltage has gotten above the float
     charger_state$ = "float"          'battery float voltage go to the charger battery float state
  ElseIf sol_watts < LOW_SOL_WATTS Then  'else if the input watts is less than low solar watts
     pwm = PWM_MAX                     'it means there is not much power being generated
     set_pwm_duty                      'set the pwm = 100% so we get as much power as possible
                                       'and stay in the charger on state
  Else
     pwm = bat_v * 100 / SOL_V + 5     'else if we are making more than low solar watts figure
     set_pwm_duty                      'out what the pwm value should be
     charger_state$ = "bulk"           'and change the charger state
  EndIf
'---------
ElseIf charger_state$="bulk" Then
  If sol_watts < MIN_SOL_WATTS Then  'if watts input is less than
      charger_state$ = "off"          'the minimum solar watts then it is getting dark so
      off_count = OFF_NUM             'go to the charger off state
      Pin(PWM_ENABLE_PIN)=0
  ElseIf bat_v > BAT_FLT_V Then      'if battery voltage is now above the
      charger_state$ = "float"       'float voltage change to float state
      DELTA= Abs(DELTA)
      pwm = pwm-5  'start float state with pwm a bit lower
      set_pwm_duty
  ElseIf sol_watts < LOW_SOL_WATTS Then  'if the solar input watts is less than
      charger_state$ = "on"              'low solar watts then there is not much
                                         'power being made by the solar panel
      Pin(PWM_ENABLE_PIN)=1              'so go to charger on state

'this is where we do the Peak Power Tracking or Maximum Power algorithm
  Else
      If (old_sol_watts >= sol_watts) Then
        DELTA = -DELTA        ' DELTA to make pwm increase or decrease to maximize watts
      EndIf
      pwm = pwm + DELTA          ' add DELTA to change PWM duty cycle for PPT algorithm
      If pwm > 94 Then pwm = 94  ' max limit for bulk state
      old_sol_watts = sol_watts  ' load old_watts with current watts value for next time
      set_pwm_duty               ' set pwm duty cycle to pwm value
  EndIf
'---------
ElseIf charger_state$="float" Then
    If sol_watts < MIN_SOL_WATTS Then  'if watts input from the solar panel is less than
      charger_state$ = "off"          'the minimum solar watts then it is getting dark so
      off_count = OFF_NUM             'go to the charger off state
      Pin(PWM_ENABLE_PIN)=0
    Else
      CH = PWM * (1 - bat_v / BAT_FLT_V)    ' PWM NEW=PWM OLD + CH
      If CH > 0.1 Then CH=0.1     'MAX POSITIVE CHANGE. PREVENT voltage OVERSHOOT
      PWM = PWM + CH
'      PWM = PWM * BAT_FLT_V / bat_v   'CHANGE IS PROPORTIONAL TO VOLTAGE ERROR
      If (pwm >= PWM_MAX) Then    'if pwm gets up to 100 it means we can't keep the batt at
          pwm = 97
          charger_state$ = "bulk"  'float voltage so jump to bulk state to charge the battery
      EndIf
    set_pwm_duty
    EndIf
ElseIf charger_state$="off" Then
    'when we jump into the charger off state, off_count is set with OFF_NUM
  If off_count > 0 Then
    off_count = off_count - 1
' this is to allow the battery voltage to settle down to see if the
' battery has been disconnected
  ElseIf (bat_v>HIGH_BAT_V) And (bat_v<BAT_FLT_V) And (SOL_V>bat_v) Then
      charger_state$ = "float"       'if batt voltage is still high and solar volts are high
      pwm = 100 * BAT_FLT_V / SOL_V
      set_pwm_duty                   'change charger state to battery float
      Pin(PWM_ENABLE_PIN)=1
      Timer=0   'this signals the main loop to pause 1.5 seconds
  ElseIf (bat_v>MIN_BAT_V) And (bat_v<BAT_FLT_V) And (SOL_V>bat_v) Then
      pwm = bat_v * 100 / SOL_V + 5
'if battery volts aren't quite so high but we have solar volts
      set_pwm_duty              'greater than battery volts showing it is day light then
      charger_state$ = "on"     'change charger state to on so we start charging
      Pin(PWM_ENABLE_PIN)=1
      Timer=0   'this signals the main loop to pause 1.5 seconds
  EndIf                              'else stay in the off state
Else    'if charger state is not one of the above, error,turn charger off
      Pin(PWM_ENABLE_PIN)=0
EndIf
End Sub

'----------end of program--------------

reboot: 'arrive here if bat voltage has gone too high
Chdir "b:\"
Open "diagnose.txt" For random As #1
Seek #1,10 : ovr = Val(Input$(5,#1))
ovr=ovr+1: Seek #1,10          ' add 1 and then save it
Print #1,Format$(ovr,"%5.0fA")
Close #1            ' close the diagnose file
Run "autorun.bas"   ' restart autorun


'******************* restart.bas *****************
'open the file, add 1 to the wdr value, close the file, run autorun
Chdir "b:\"
Open "diagnose.txt" For random As #1
Seek #1,1 : wdr = Val(Input$(5,#1))
wdr=wdr+1: Seek #1,1
Print #1,Format$(wdr,"%5.0f")
Close #1
Run "autorun.bas"
